from __future__ import annotations

from typing import (
    TypeVar,
    get_args,
    get_origin,
    get_type_hints,
)

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    Device,
    Signal,
)
from ophyd_async.tango.signal import (
    TangoSignalBackend,
    __tango_signal_auto,
    make_backend,
)
from tango import DeviceProxy as DeviceProxy
from tango.asyncio import DeviceProxy as AsyncDeviceProxy

T = TypeVar("T")


class TangoDevice(Device):
    """
    General class for TangoDevices. Extends Device to provide attributes for Tango
    devices.

    Parameters
    ----------
    trl: str
        Tango resource locator, typically of the device server.
    device_proxy: Optional[Union[AsyncDeviceProxy, SyncDeviceProxy]]
        Asynchronous or synchronous DeviceProxy object for the device. If not provided,
        an asynchronous DeviceProxy object will be created using the trl and awaited
        when the device is connected.
    """

    trl: str = ""
    proxy: DeviceProxy | None = None
    _polling: tuple[bool, float, float | None, float | None] = (False, 0.1, None, 0.1)
    _signal_polling: dict[str, tuple[bool, float, float, float]] = {}
    _poll_only_annotated_signals: bool = True

    def __init__(
        self,
        trl: str | None = None,
        device_proxy: DeviceProxy | None = None,
        name: str = "",
    ) -> None:
        self.trl = trl if trl else ""
        self.proxy = device_proxy
        tango_create_children_from_annotations(self)
        super().__init__(name=name)

    def set_trl(self, trl: str):
        """Set the Tango resource locator."""
        if not isinstance(trl, str):
            raise ValueError("TRL must be a string.")
        self.trl = trl

    async def connect(
        self,
        mock: bool = False,
        timeout: float = DEFAULT_TIMEOUT,
        force_reconnect: bool = False,
    ):
        if self.trl and self.proxy is None:
            self.proxy = await AsyncDeviceProxy(self.trl)
        elif self.proxy and not self.trl:
            self.trl = self.proxy.name()

        # Set the trl of the signal backends
        for child in self.children():
            if isinstance(child[1], Signal):
                if isinstance(child[1]._backend, TangoSignalBackend):  # noqa: SLF001
                    resource_name = child[0].lstrip("_")
                    read_trl = f"{self.trl}/{resource_name}"
                    child[1]._backend.set_trl(read_trl, read_trl)  # noqa: SLF001

        if self.proxy is not None:
            self.register_signals()
            await _fill_proxy_entries(self)

        # set_name should be called again to propagate the new signal names
        self.set_name(self.name)

        # Set the polling configuration
        if self._polling[0]:
            for child in self.children():
                child_type = type(child[1])
                if issubclass(child_type, Signal):
                    if isinstance(child[1]._backend, TangoSignalBackend):  # noqa: SLF001  # type: ignore
                        child[1]._backend.set_polling(*self._polling)  # noqa: SLF001  # type: ignore
                        child[1]._backend.allow_events(False)  # noqa: SLF001  # type: ignore
        if self._signal_polling:
            for signal_name, polling in self._signal_polling.items():
                if hasattr(self, signal_name):
                    attr = getattr(self, signal_name)
                    if isinstance(attr._backend, TangoSignalBackend):  # noqa: SLF001
                        attr._backend.set_polling(*polling)  # noqa: SLF001
                        attr._backend.allow_events(False)  # noqa: SLF001

        await super().connect(mock=mock, timeout=timeout)

    # Users can override this method to register new signals
    def register_signals(self):
        pass


def tango_polling(
    polling: tuple[float, float, float]
    | dict[str, tuple[float, float, float]]
    | None = None,
    signal_polling: dict[str, tuple[float, float, float]] | None = None,
):
    """
    Class decorator to configure polling for Tango devices.

    This decorator allows for the configuration of both device-level and signal-level
    polling for Tango devices. Polling is useful for device servers that do not support
    event-driven updates.

    Parameters
    ----------
    polling : Optional[Union[Tuple[float, float, float],
              Dict[str, Tuple[float, float, float]]]], optional
        Device-level polling configuration as a tuple of three floats representing the
        polling interval, polling timeout, and polling delay. Alternatively,
        a dictionary can be provided to specify signal-level polling configurations
        directly.
    signal_polling : Optional[Dict[str, Tuple[float, float, float]]], optional
        Signal-level polling configuration as a dictionary where keys are signal names
        and values are tuples of three floats representing the polling interval, polling
        timeout, and polling delay.
    """
    if isinstance(polling, dict):
        signal_polling = polling
        polling = None

    def decorator(cls):
        if polling is not None:
            cls._polling = (True, *polling)
        if signal_polling is not None:
            cls._signal_polling = {k: (True, *v) for k, v in signal_polling.items()}
        return cls

    return decorator


def tango_create_children_from_annotations(
    device: TangoDevice, included_optional_fields: tuple[str, ...] = ()
):
    """Initialize blocks at __init__ of `device`."""
    for name, device_type in get_type_hints(type(device)).items():
        if name in ("_name", "parent"):
            continue

        # device_type, is_optional = _strip_union(device_type)
        # if is_optional and name not in included_optional_fields:
        #     continue
        #
        # is_device_vector, device_type = _strip_device_vector(device_type)
        # if is_device_vector:
        #     n_device_vector = DeviceVector()
        #     setattr(device, name, n_device_vector)

        # else:
        origin = get_origin(device_type)
        origin = origin if origin else device_type

        if issubclass(origin, Signal):
            type_args = get_args(device_type)
            datatype = type_args[0] if type_args else None
            backend = make_backend(datatype=datatype, device_proxy=device.proxy)
            setattr(device, name, origin(name=name, backend=backend))

        elif issubclass(origin, Device) or isinstance(origin, Device):
            assert callable(origin), f"{origin} is not callable."
            setattr(device, name, origin())


async def _fill_proxy_entries(device: TangoDevice):
    if device.proxy is None:
        raise RuntimeError(f"Device proxy is not connected for {device.name}")
    proxy_trl = device.trl
    children = [name.lstrip("_") for name, _ in device.children()]
    proxy_attributes = list(device.proxy.get_attribute_list())
    proxy_commands = list(device.proxy.get_command_list())
    combined = proxy_attributes + proxy_commands

    for name in combined:
        if name not in children:
            full_trl = f"{proxy_trl}/{name}"
            try:
                auto_signal = await __tango_signal_auto(
                    trl=full_trl, device_proxy=device.proxy
                )
                setattr(device, name, auto_signal)
            except RuntimeError as e:
                if "Commands with different in and out dtypes" in str(e):
                    print(
                        f"Skipping {name}. Commands with different in and out dtypes"
                        f" are not supported."
                    )
                    continue
                raise e


# def _strip_union(field: T | T) -> tuple[T, bool]:
#     if get_origin(field) is Union:
#         args = get_args(field)
#         is_optional = type(None) in args
#         for arg in args:
#             if arg is not type(None):
#                 return arg, is_optional
#     return field, False
#
#
# def _strip_device_vector(field: type[Device]) -> tuple[bool, type[Device]]:
#     if get_origin(field) is DeviceVector:
#         return True, get_args(field)[0]
#     return False, field



import asyncio

from ophyd_async.core import (
    DetectorController,
    DetectorTrigger,
    wait_for_value,
)
from ophyd_async.core._detector import TriggerInfo
from ophyd_async.core._status import AsyncStatus

from ._block import PcapBlock


class PandaPcapController(DetectorController):
    def __init__(self, pcap: PcapBlock) -> None:
        self.pcap = pcap
        self._arm_status: AsyncStatus | None = None

    def get_deadtime(self, exposure: float | None) -> float:
        return 0.000000008

    async def prepare(self, trigger_info: TriggerInfo):
        assert trigger_info.trigger in (
            DetectorTrigger.constant_gate,
            DetectorTrigger.variable_gate,
        ), "Only constant_gate and variable_gate triggering is supported on the PandA"

    async def arm(self):
        self._arm_status = self.pcap.arm.set(True)
        await wait_for_value(self.pcap.active, True, timeout=1)

    async def wait_for_idle(self):
        pass

    async def disarm(self):
        await asyncio.gather(self.pcap.arm.set(False))
        await wait_for_value(self.pcap.active, False, timeout=1)



from pathlib import Path
from unittest.mock import ANY, MagicMock

import pytest

from ophyd_async.core import DeviceCollector, get_mock_put, set_mock_value
from ophyd_async.epics.eiger._odin_io import Odin, OdinWriter, Writing

OdinDriverAndWriter = tuple[Odin, OdinWriter]


@pytest.fixture
def odin_driver_and_writer(RE) -> OdinDriverAndWriter:
    with DeviceCollector(mock=True):
        driver = Odin("")
        writer = OdinWriter(MagicMock(), lambda: "odin", driver)
    return driver, writer


async def test_when_open_called_then_file_correctly_set(
    odin_driver_and_writer: OdinDriverAndWriter,
):
    driver, writer = odin_driver_and_writer
    path_info = writer._path_provider.return_value
    expected_path = "/tmp"
    expected_filename = "filename.h5"
    path_info.directory_path = Path(expected_path)
    path_info.filename = expected_filename

    await writer.open()

    get_mock_put(driver.file_path).assert_called_once_with(
        expected_path, wait=ANY, timeout=ANY
    )
    get_mock_put(driver.file_name).assert_called_once_with(
        expected_filename, wait=ANY, timeout=ANY
    )


async def test_when_open_called_then_all_expected_signals_set(
    odin_driver_and_writer: OdinDriverAndWriter,
):
    driver, writer = odin_driver_and_writer
    await writer.open()

    get_mock_put(driver.data_type).assert_called_once_with(
        "uint16", wait=ANY, timeout=ANY
    )
    get_mock_put(driver.num_to_capture).assert_called_once_with(
        0, wait=ANY, timeout=ANY
    )

    get_mock_put(driver.capture).assert_called_once_with(
        Writing.ON, wait=ANY, timeout=ANY
    )


async def test_given_data_shape_set_when_open_called_then_describe_has_correct_shape(
    odin_driver_and_writer: OdinDriverAndWriter,
):
    driver, writer = odin_driver_and_writer
    set_mock_value(driver.image_width, 1024)
    set_mock_value(driver.image_height, 768)
    description = await writer.open()
    assert description["data"]["shape"] == [768, 1024]


async def test_when_closed_then_data_capture_turned_off(
    odin_driver_and_writer: OdinDriverAndWriter,
):
    driver, writer = odin_driver_and_writer
    await writer.close()
    get_mock_put(driver.capture).assert_called_once_with(
        Writing.OFF, wait=ANY, timeout=ANY
    )



from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    AsyncStatus,
    ConfigSignal,
    HintedSignal,
    SignalR,
    SignalRW,
    SignalX,
)
from ophyd_async.tango import TangoReadable, tango_polling


# Enable device level polling, useful for servers that do not support events
# Polling for individual signal can be enabled with a dict
@tango_polling({"counts": (1.0, 0.1, 0.1), "sample_time": (0.1, 0.1, 0.1)})
class TangoCounter(TangoReadable):
    # Enter the name and type of the signals you want to use
    # If type is None or Signal, the type will be inferred from the Tango device
    counts: SignalR[int]
    sample_time: SignalRW[float]
    start: SignalX
    _reset: SignalX

    def __init__(self, trl: str | None = "", name=""):
        super().__init__(trl, name=name)
        self.add_readables([self.counts], HintedSignal)
        self.add_readables([self.sample_time], ConfigSignal)

    @AsyncStatus.wrap
    async def trigger(self) -> None:
        sample_time = await self.sample_time.get_value()
        timeout = sample_time + DEFAULT_TIMEOUT
        await self.start.trigger(wait=True, timeout=timeout)

    @AsyncStatus.wrap
    async def reset(self) -> None:
        await self._reset.trigger(wait=True, timeout=DEFAULT_TIMEOUT)



"""EPICS Signals over CA or PVA"""

from __future__ import annotations

from enum import Enum


def _make_unavailable_class(error: Exception) -> type:
    class TransportNotAvailable:
        def __init__(*args, **kwargs):
            raise NotImplementedError("Transport not available") from error

    return TransportNotAvailable


try:
    from ._aioca import CaSignalBackend
except ImportError as ca_error:
    CaSignalBackend = _make_unavailable_class(ca_error)


try:
    from ._p4p import PvaSignalBackend
except ImportError as pva_error:
    PvaSignalBackend = _make_unavailable_class(pva_error)


class _EpicsTransport(Enum):
    """The sorts of transport EPICS support"""

    #: Use Channel Access (using aioca library)
    ca = CaSignalBackend
    #: Use PVAccess (using p4p library)
    pva = PvaSignalBackend



import inspect
from enum import Enum

from typing_extensions import TypedDict

from ophyd_async.core import RuntimeSubsetEnum

common_meta = {
    "units",
    "precision",
}


class LimitPair(TypedDict):
    high: float | None
    low: float | None


class Limits(TypedDict):
    alarm: LimitPair
    control: LimitPair
    display: LimitPair
    warning: LimitPair


def get_supported_values(
    pv: str,
    datatype: type[str] | None,
    pv_choices: tuple[str, ...],
) -> dict[str, str]:
    if inspect.isclass(datatype) and issubclass(datatype, RuntimeSubsetEnum):
        if not set(datatype.choices).issubset(set(pv_choices)):
            raise TypeError(
                f"{pv} has choices {pv_choices}, "
                f"which is not a superset of {str(datatype)}."
            )
        return {x: x or "_" for x in pv_choices}
    elif inspect.isclass(datatype) and issubclass(datatype, Enum):
        if not issubclass(datatype, str):
            raise TypeError(
                f"{pv} is type Enum but {datatype} does not inherit from String."
            )

        choices = tuple(v.value for v in datatype)
        if set(choices) != set(pv_choices):
            raise TypeError(
                f"{pv} has choices {pv_choices}, "
                f"which do not match {datatype}, which has {choices}."
            )
        return {x: datatype(x) if x else "_" for x in pv_choices}
    elif datatype is None or datatype is str:
        return {x: x or "_" for x in pv_choices}

    raise TypeError(
        f"{pv} has choices {pv_choices}. "
        "Use an Enum or SubsetEnum to represent this."
    )



import inspect
import logging
import sys
from collections.abc import Sequence
from dataclasses import dataclass
from enum import Enum
from math import isnan, nan
from typing import Any, get_origin

import numpy as np
from aioca import (
    FORMAT_CTRL,
    FORMAT_RAW,
    FORMAT_TIME,
    CANothing,
    Subscription,
    caget,
    camonitor,
    caput,
)
from aioca.types import AugmentedValue, Dbr, Format
from bluesky.protocols import Reading
from epicscorelibs.ca import dbr
from event_model import DataKey
from event_model.documents.event_descriptor import Dtype

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    NotConnected,
    ReadingValueCallback,
    RuntimeSubsetEnum,
    SignalBackend,
    T,
    get_dtype,
    get_unique,
    wait_for_connection,
)

from ._common import LimitPair, Limits, common_meta, get_supported_values

dbr_to_dtype: dict[Dbr, Dtype] = {
    dbr.DBR_STRING: "string",
    dbr.DBR_SHORT: "integer",
    dbr.DBR_FLOAT: "number",
    dbr.DBR_CHAR: "string",
    dbr.DBR_LONG: "integer",
    dbr.DBR_DOUBLE: "number",
}


def _data_key_from_augmented_value(
    value: AugmentedValue,
    *,
    choices: list[str] | None = None,
    dtype: Dtype | None = None,
) -> DataKey:
    """Use the return value of get with FORMAT_CTRL to construct a DataKey
    describing the signal. See docstring of AugmentedValue for expected
    value fields by DBR type.

    Args:
        value (AugmentedValue): Description of the the return type of a DB record
        choices: Optional list of enum choices to pass as metadata in the datakey
        dtype: Optional override dtype when AugmentedValue is ambiguous, e.g. booleans

    Returns:
        DataKey: A rich DataKey describing the DB record
    """
    source = f"ca://{value.name}"
    assert value.ok, f"Error reading {source}: {value}"

    scalar = value.element_count == 1
    dtype = dtype or dbr_to_dtype[value.datatype]  # type: ignore

    dtype_numpy = np.dtype(dbr.DbrCodeToType[value.datatype].dtype).descr[0][1]

    d = DataKey(
        source=source,
        dtype=dtype if scalar else "array",
        # Ignore until https://github.com/bluesky/event-model/issues/308
        dtype_numpy=dtype_numpy,  # type: ignore
        # strictly value.element_count >= len(value)
        shape=[] if scalar else [len(value)],
    )
    for key in common_meta:
        attr = getattr(value, key, nan)
        if isinstance(attr, str) or not isnan(attr):
            d[key] = attr

    if choices is not None:
        d["choices"] = choices  # type: ignore

    if limits := _limits_from_augmented_value(value):
        d["limits"] = limits  # type: ignore

    return d


def _limits_from_augmented_value(value: AugmentedValue) -> Limits:
    def get_limits(limit: str) -> LimitPair:
        low = getattr(value, f"lower_{limit}_limit", nan)
        high = getattr(value, f"upper_{limit}_limit", nan)
        return LimitPair(
            low=None if isnan(low) else low, high=None if isnan(high) else high
        )

    return Limits(
        alarm=get_limits("alarm"),
        control=get_limits("ctrl"),
        display=get_limits("disp"),
        warning=get_limits("warning"),
    )


@dataclass
class CaConverter:
    read_dbr: Dbr | None
    write_dbr: Dbr | None

    def write_value(self, value) -> Any:
        return value

    def value(self, value: AugmentedValue):
        # for channel access ca_xxx classes, this
        # invokes __pos__ operator to return an instance of
        # the builtin base class
        return +value  # type: ignore

    def reading(self, value: AugmentedValue) -> Reading:
        return {
            "value": self.value(value),
            "timestamp": value.timestamp,
            "alarm_severity": -1 if value.severity > 2 else value.severity,
        }

    def get_datakey(self, value: AugmentedValue) -> DataKey:
        return _data_key_from_augmented_value(value)


class CaLongStrConverter(CaConverter):
    def __init__(self):
        return super().__init__(dbr.DBR_CHAR_STR, dbr.DBR_CHAR_STR)

    def write_value(self, value: str):
        # Add a null in here as this is what the commandline caput does
        # TODO: this should be in the server so check if it can be pushed to asyn
        return value + "\0"


class CaArrayConverter(CaConverter):
    def value(self, value: AugmentedValue):
        return np.array(value, copy=False)


@dataclass
class CaEnumConverter(CaConverter):
    """To prevent issues when a signal is restarted and returns with different enum
    values or orders, we put treat an Enum signal as a string, and cache the
    choices on this class.
    """

    choices: dict[str, str]

    def write_value(self, value: Enum | str):
        if isinstance(value, Enum):
            return value.value
        else:
            return value

    def value(self, value: AugmentedValue):
        return self.choices[value]  # type: ignore

    def get_datakey(self, value: AugmentedValue) -> DataKey:
        # Sometimes DBR_TYPE returns as String, must pass choices still
        return _data_key_from_augmented_value(value, choices=list(self.choices.keys()))


@dataclass
class CaBoolConverter(CaConverter):
    def value(self, value: AugmentedValue) -> bool:
        return bool(value)

    def get_datakey(self, value: AugmentedValue) -> DataKey:
        return _data_key_from_augmented_value(value, dtype="boolean")


class DisconnectedCaConverter(CaConverter):
    def __getattribute__(self, __name: str) -> Any:
        raise NotImplementedError("No PV has been set as connect() has not been called")


def make_converter(
    datatype: type | None, values: dict[str, AugmentedValue]
) -> CaConverter:
    pv = list(values)[0]
    pv_dbr = get_unique({k: v.datatype for k, v in values.items()}, "datatypes")
    is_array = bool([v for v in values.values() if v.element_count > 1])
    if is_array and datatype is str and pv_dbr == dbr.DBR_CHAR:
        # Override waveform of chars to be treated as string
        return CaLongStrConverter()
    elif is_array and pv_dbr == dbr.DBR_STRING:
        # Waveform of strings, check we wanted this
        if datatype:
            datatype_dtype = get_dtype(datatype)
            if not datatype_dtype or not np.can_cast(datatype_dtype, np.str_):
                raise TypeError(f"{pv} has type [str] not {datatype.__name__}")
        return CaArrayConverter(pv_dbr, None)
    elif is_array:
        pv_dtype = get_unique({k: v.dtype for k, v in values.items()}, "dtypes")  # type: ignore
        # This is an array
        if datatype:
            # Check we wanted an array of this type
            dtype = get_dtype(datatype)
            if not dtype:
                raise TypeError(f"{pv} has type [{pv_dtype}] not {datatype.__name__}")
            if dtype != pv_dtype:
                raise TypeError(f"{pv} has type [{pv_dtype}] not [{dtype}]")
        return CaArrayConverter(pv_dbr, None)  # type: ignore
    elif pv_dbr == dbr.DBR_ENUM and datatype is bool:
        # Database can't do bools, so are often representated as enums,
        # CA can do int
        pv_choices_len = get_unique(
            {k: len(v.enums) for k, v in values.items()}, "number of choices"
        )
        if pv_choices_len != 2:
            raise TypeError(f"{pv} has {pv_choices_len} choices, can't map to bool")
        return CaBoolConverter(dbr.DBR_SHORT, dbr.DBR_SHORT)
    elif pv_dbr == dbr.DBR_ENUM:
        # This is an Enum
        pv_choices = get_unique(
            {k: tuple(v.enums) for k, v in values.items()}, "choices"
        )
        supported_values = get_supported_values(pv, datatype, pv_choices)
        return CaEnumConverter(dbr.DBR_STRING, None, supported_values)
    else:
        value = list(values.values())[0]
        # Done the dbr check, so enough to check one of the values
        if datatype and not isinstance(value, datatype):
            # Allow int signals to represent float records when prec is 0
            is_prec_zero_float = (
                isinstance(value, float)
                and get_unique({k: v.precision for k, v in values.items()}, "precision")
                == 0
            )
            if not (datatype is int and is_prec_zero_float):
                raise TypeError(
                    f"{pv} has type {type(value).__name__.replace('ca_', '')} "
                    + f"not {datatype.__name__}"
                )
    return CaConverter(pv_dbr, None)  # type: ignore


_tried_pyepics = False


def _use_pyepics_context_if_imported():
    global _tried_pyepics
    if not _tried_pyepics:
        ca = sys.modules.get("epics.ca", None)
        if ca:
            ca.use_initial_context()
        _tried_pyepics = True


class CaSignalBackend(SignalBackend[T]):
    _ALLOWED_DATATYPES = (
        bool,
        int,
        float,
        str,
        Sequence,
        Enum,
        RuntimeSubsetEnum,
        np.ndarray,
    )

    @classmethod
    def datatype_allowed(cls, dtype: Any) -> bool:
        stripped_origin = get_origin(dtype) or dtype
        if dtype is None:
            return True

        return inspect.isclass(stripped_origin) and issubclass(
            stripped_origin, cls._ALLOWED_DATATYPES
        )

    def __init__(self, datatype: type[T] | None, read_pv: str, write_pv: str):
        self.datatype = datatype
        if not CaSignalBackend.datatype_allowed(self.datatype):
            raise TypeError(f"Given datatype {self.datatype} unsupported in CA.")
        self.read_pv = read_pv
        self.write_pv = write_pv
        self.initial_values: dict[str, AugmentedValue] = {}
        self.converter: CaConverter = DisconnectedCaConverter(None, None)
        self.subscription: Subscription | None = None

    def source(self, name: str):
        return f"ca://{self.read_pv}"

    async def _store_initial_value(self, pv, timeout: float = DEFAULT_TIMEOUT):
        try:
            self.initial_values[pv] = await caget(
                pv, format=FORMAT_CTRL, timeout=timeout
            )
        except CANothing as exc:
            logging.debug(f"signal ca://{pv} timed out")
            raise NotConnected(f"ca://{pv}") from exc

    async def connect(self, timeout: float = DEFAULT_TIMEOUT):
        _use_pyepics_context_if_imported()
        if self.read_pv != self.write_pv:
            # Different, need to connect both
            await wait_for_connection(
                read_pv=self._store_initial_value(self.read_pv, timeout=timeout),
                write_pv=self._store_initial_value(self.write_pv, timeout=timeout),
            )
        else:
            # The same, so only need to connect one
            await self._store_initial_value(self.read_pv, timeout=timeout)
        self.converter = make_converter(self.datatype, self.initial_values)

    async def put(self, value: T | None, wait=True, timeout=None):
        if value is None:
            write_value = self.initial_values[self.write_pv]
        else:
            write_value = self.converter.write_value(value)
        await caput(
            self.write_pv,
            write_value,
            datatype=self.converter.write_dbr,
            wait=wait,
            timeout=timeout,
        )

    async def _caget(self, format: Format) -> AugmentedValue:
        return await caget(
            self.read_pv,
            datatype=self.converter.read_dbr,
            format=format,
            timeout=None,
        )

    async def get_datakey(self, source: str) -> DataKey:
        value = await self._caget(FORMAT_CTRL)
        return self.converter.get_datakey(value)

    async def get_reading(self) -> Reading:
        value = await self._caget(FORMAT_TIME)
        return self.converter.reading(value)

    async def get_value(self) -> T:
        value = await self._caget(FORMAT_RAW)
        return self.converter.value(value)

    async def get_setpoint(self) -> T:
        value = await caget(
            self.write_pv,
            datatype=self.converter.read_dbr,
            format=FORMAT_RAW,
            timeout=None,
        )
        return self.converter.value(value)

    def set_callback(self, callback: ReadingValueCallback[T] | None) -> None:
        if callback:
            assert (
                not self.subscription
            ), "Cannot set a callback when one is already set"
            self.subscription = camonitor(
                self.read_pv,
                lambda v: callback(self.converter.reading(v), self.converter.value(v)),
                datatype=self.converter.read_dbr,
                format=FORMAT_TIME,
            )
        else:
            if self.subscription:
                self.subscription.close()
            self.subscription = None



from enum import Enum
from typing import TypeVar, get_args, get_origin

import numpy as np
from pydantic import BaseModel, ConfigDict, model_validator

TableSubclass = TypeVar("TableSubclass", bound="Table")


def _concat(value1, value2):
    if isinstance(value1, np.ndarray):
        return np.concatenate((value1, value2))
    else:
        return value1 + value2


class Table(BaseModel):
    """An abstraction of a Table of str to numpy array."""

    model_config = ConfigDict(validate_assignment=True, strict=False)

    @staticmethod
    def row(cls: type[TableSubclass], **kwargs) -> TableSubclass:  # type: ignore
        arrayified_kwargs = {}
        for field_name, field_value in cls.model_fields.items():
            value = kwargs.pop(field_name)
            if field_value.default_factory is None:
                raise ValueError(
                    "`Table` models should have default factories for their "
                    "mutable empty columns."
                )
            default_array = field_value.default_factory()
            if isinstance(default_array, np.ndarray):
                arrayified_kwargs[field_name] = np.array(
                    [value], dtype=default_array.dtype
                )
            elif issubclass(type(value), Enum) and isinstance(value, str):
                arrayified_kwargs[field_name] = [value]
            else:
                raise TypeError(
                    "Row column should be numpy arrays or sequence of string `Enum`."
                )
        if kwargs:
            raise TypeError(
                f"Unexpected keyword arguments {kwargs.keys()} for {cls.__name__}."
            )
        return cls(**arrayified_kwargs)

    def __add__(self, right: TableSubclass) -> TableSubclass:
        """Concatenate the arrays in field values."""

        if type(right) is not type(self):
            raise RuntimeError(
                f"{right} is not a `Table`, or is not the same "
                f"type of `Table` as {self}."
            )

        return type(right)(
            **{
                field_name: _concat(
                    getattr(self, field_name), getattr(right, field_name)
                )
                for field_name in self.model_fields
            }
        )

    def numpy_dtype(self) -> np.dtype:
        dtype = []
        for field_name, field_value in self.model_fields.items():
            if np.ndarray in (
                get_origin(field_value.annotation),
                field_value.annotation,
            ):
                dtype.append((field_name, getattr(self, field_name).dtype))
            else:
                enum_type = get_args(field_value.annotation)[0]
                assert issubclass(enum_type, Enum)
                enum_values = [element.value for element in enum_type]
                max_length_in_enum = max(len(value) for value in enum_values)
                dtype.append((field_name, np.dtype(f"<U{max_length_in_enum}")))

        return np.dtype(dtype)

    def numpy_table(self):
        # It would be nice to be able to use np.transpose for this,
        # but it defaults to the largest dtype for everything.
        dtype = self.numpy_dtype()
        transposed_list = [
            np.array(tuple(row), dtype=dtype)
            for row in zip(*self.numpy_columns(), strict=False)
        ]
        transposed = np.array(transposed_list, dtype=dtype)
        return transposed

    def numpy_columns(self) -> list[np.ndarray]:
        """Columns in the table can be lists of string enums or numpy arrays.

        This method returns the columns, converting the string enums to numpy arrays.
        """

        columns = []
        for field_name, field_value in self.model_fields.items():
            if np.ndarray in (
                get_origin(field_value.annotation),
                field_value.annotation,
            ):
                columns.append(getattr(self, field_name))
            else:
                enum_type = get_args(field_value.annotation)[0]
                assert issubclass(enum_type, Enum)
                enum_values = [element.value for element in enum_type]
                max_length_in_enum = max(len(value) for value in enum_values)
                dtype = np.dtype(f"<U{max_length_in_enum}")

                columns.append(
                    np.array(
                        [enum.value for enum in getattr(self, field_name)], dtype=dtype
                    )
                )

        return columns

    @model_validator(mode="after")
    def validate_arrays(self) -> "Table":
        first_length = len(next(iter(self))[1])
        assert all(
            len(field_value) == first_length for _, field_value in self
        ), "Rows should all be of equal size."

        if not all(
            # Checks if the values are numpy subtypes if the array is a numpy array,
            # or if the value is a string enum.
            np.issubdtype(getattr(self, field_name).dtype, default_array.dtype)
            if isinstance(
                default_array := self.model_fields[field_name].default_factory(),  # type: ignore
                np.ndarray,
            )
            else issubclass(get_args(field_value.annotation)[0], Enum)
            for field_name, field_value in self.model_fields.items()
        ):
            raise ValueError(
                f"Cannot construct a `{type(self).__name__}`, "
                "some rows have incorrect types."
            )

        return self



from enum import Enum

from ophyd_async.core import ConfigSignal, DeviceVector, HintedSignal, StandardReadable
from ophyd_async.epics.signal import epics_signal_r, epics_signal_rw


class EnergyMode(str, Enum):
    """Energy mode for `Sensor`"""

    #: Low energy mode
    low = "Low Energy"
    #: High energy mode
    high = "High Energy"


class Sensor(StandardReadable):
    """A demo sensor that produces a scalar value based on X and Y Movers"""

    def __init__(self, prefix: str, name="") -> None:
        # Define some signals
        with self.add_children_as_readables(HintedSignal):
            self.value = epics_signal_r(float, prefix + "Value")
        with self.add_children_as_readables(ConfigSignal):
            self.mode = epics_signal_rw(EnergyMode, prefix + "Mode")

        super().__init__(name=name)


class SensorGroup(StandardReadable):
    def __init__(self, prefix: str, name: str = "", sensor_count: int = 3) -> None:
        with self.add_children_as_readables():
            self.sensors = DeviceVector(
                {i: Sensor(f"{prefix}{i}:") for i in range(1, sensor_count + 1)}
            )

        super().__init__(name)



import asyncio

import numpy as np
import pytest
from pydantic import ValidationError

from ophyd_async.core import DEFAULT_TIMEOUT, DeviceCollector, set_mock_value
from ophyd_async.epics.pvi import fill_pvi_entries
from ophyd_async.fastcs.panda import (
    CommonPandaBlocks,
    PcompInfo,
    SeqTable,
    SeqTableInfo,
    StaticPcompTriggerLogic,
    StaticSeqTableTriggerLogic,
)
from ophyd_async.fastcs.panda._table import SeqTrigger


@pytest.fixture
async def mock_panda():
    class Panda(CommonPandaBlocks):
        def __init__(self, prefix: str, name: str = ""):
            self._prefix = prefix
            super().__init__(name)

        async def connect(self, mock: bool = False, timeout: float = DEFAULT_TIMEOUT):
            await fill_pvi_entries(
                self, self._prefix + "PVI", timeout=timeout, mock=mock
            )
            await super().connect(mock=mock, timeout=timeout)

    async with DeviceCollector(mock=True):
        mock_panda = Panda("PANDAQSRV:", "mock_panda")

    assert mock_panda.name == "mock_panda"
    return mock_panda


async def test_seq_table_trigger_logic(mock_panda):
    trigger_logic = StaticSeqTableTriggerLogic(mock_panda.seq[1])
    seq_table = (
        SeqTable.row(outa1=True, outa2=True)
        + SeqTable.row(outa1=False, outa2=False)
        + SeqTable.row(outa1=True, outa2=False)
        + SeqTable.row(outa1=False, outa2=True)
    )
    seq_table_info = SeqTableInfo(sequence_table=seq_table, repeats=1)

    async def set_active(value: bool):
        await asyncio.sleep(0.1)
        set_mock_value(mock_panda.seq[1].active, value)

    await trigger_logic.prepare(seq_table_info)
    await asyncio.gather(trigger_logic.kickoff(), set_active(True))
    await asyncio.gather(trigger_logic.complete(), set_active(False))


async def test_pcomp_trigger_logic(mock_panda):
    trigger_logic = StaticPcompTriggerLogic(mock_panda.pcomp[1])
    pcomp_info = PcompInfo(
        start_postion=0,
        pulse_width=1,
        rising_edge_step=1,
        number_of_pulses=5,
        direction="Positive",
    )

    async def set_active(value: bool):
        await asyncio.sleep(0.1)
        set_mock_value(mock_panda.pcomp[1].active, value)

    await trigger_logic.prepare(pcomp_info)
    await asyncio.gather(trigger_logic.kickoff(), set_active(True))
    await asyncio.gather(trigger_logic.complete(), set_active(False))


@pytest.mark.parametrize(
    ["kwargs", "error_msg"],
    [
        (
            {
                "sequence_table_factory": lambda: SeqTable.row(outc2=1),
                "repeats": 0,
                "prescale_as_us": -1,
            },
            "Input should be greater than or equal to 0 "
            "[type=greater_than_equal, input_value=-1, input_type=int]",
        ),
        (
            {
                "sequence_table_factory": lambda: (
                    SeqTable.row(outc2=True)
                    + SeqTable.row(outc2=False)
                    + SeqTable.row(outc2=True)
                    + SeqTable.row(outc2=False)
                ),
                "repeats": -1,
            },
            "Input should be greater than or equal to 0 "
            "[type=greater_than_equal, input_value=-1, input_type=int]",
        ),
        (
            {
                "sequence_table_factory": lambda: 1,
                "repeats": 1,
            },
            "Input should be a valid dictionary or instance of SeqTable "
            "[type=model_type, input_value=1, input_type=int]",
        ),
    ],
)
def test_malformed_seq_table_info(kwargs, error_msg):
    with pytest.raises(ValidationError) as exc:
        SeqTableInfo(sequence_table=kwargs.pop("sequence_table_factory")(), **kwargs)
    assert error_msg in str(exc.value)


def test_malformed_trigger_in_seq_table():
    def full_seq_table(trigger):
        return SeqTable(
            repeats=np.array([1], dtype=np.int32),
            trigger=trigger,
            position=np.array([1], dtype=np.int32),
            time1=np.array([1], dtype=np.int32),
            outa1=np.array([1], dtype=np.bool_),
            outb1=np.array([1], dtype=np.bool_),
            outc1=np.array([1], dtype=np.bool_),
            outd1=np.array([1], dtype=np.bool_),
            oute1=np.array([1], dtype=np.bool_),
            outf1=np.array([1], dtype=np.bool_),
            time2=np.array([1], dtype=np.int32),
            outa2=np.array([1], dtype=np.bool_),
            outb2=np.array([1], dtype=np.bool_),
            outc2=np.array([1], dtype=np.bool_),
            outd2=np.array([1], dtype=np.bool_),
            oute2=np.array([1], dtype=np.bool_),
            outf2=np.array([1], dtype=np.bool_),
        )

    for attempted_table in [
        np.array(["A"], dtype="U32"),
        np.array(["Immediate"], dtype="U32"),
        {"Immediate"},
    ]:
        with pytest.raises(ValidationError) as exc:
            full_seq_table(attempted_table)
        assert "Input should be an instance of Sequence" in str(exc)

    with pytest.raises(ValidationError) as exc:
        full_seq_table(["A"])
    assert (
        "Input should be 'Immediate', 'BITA=0', 'BITA=1', 'BITB=0', 'BITB=1', "
        "'BITC...' [type=enum, input_value='A', input_type=str]"
    ) in str(exc)

    # Pydantic is able to infer type from these
    table = full_seq_table([SeqTrigger.IMMEDIATE])
    assert table.trigger == ["Immediate"] == [SeqTrigger.IMMEDIATE]
    table = full_seq_table(["Immediate"])
    assert table.trigger == ["Immediate"] == [SeqTrigger.IMMEDIATE]



from __future__ import annotations

from collections.abc import Sequence

from ophyd_async.core import DEFAULT_TIMEOUT, PathProvider, SignalR, StandardDetector
from ophyd_async.epics.pvi import create_children_from_annotations, fill_pvi_entries

from ._block import CommonPandaBlocks
from ._control import PandaPcapController
from ._writer import PandaHDFWriter


class HDFPanda(CommonPandaBlocks, StandardDetector):
    def __init__(
        self,
        prefix: str,
        path_provider: PathProvider,
        config_sigs: Sequence[SignalR] = (),
        name: str = "",
    ):
        self._prefix = prefix

        create_children_from_annotations(self)
        controller = PandaPcapController(pcap=self.pcap)
        writer = PandaHDFWriter(
            prefix=prefix,
            path_provider=path_provider,
            name_provider=lambda: name,
            panda_data_block=self.data,
        )
        super().__init__(
            controller=controller,
            writer=writer,
            config_sigs=config_sigs,
            name=name,
        )

    async def connect(
        self,
        mock: bool = False,
        timeout: float = DEFAULT_TIMEOUT,
        force_reconnect: bool = False,
    ):
        # TODO: this doesn't support caching
        # https://github.com/bluesky/ophyd-async/issues/472
        await fill_pvi_entries(self, self._prefix + "PVI", timeout=timeout, mock=mock)
        await super().connect(
            mock=mock, timeout=timeout, force_reconnect=force_reconnect
        )



from collections.abc import Sequence
from enum import Enum
from typing import Annotated

import numpy as np
import numpy.typing as npt
from pydantic import Field, model_validator
from pydantic_numpy.helper.annotation import NpArrayPydanticAnnotation
from typing_extensions import TypedDict

from ophyd_async.core import Table


class PandaHdf5DatasetType(str, Enum):
    FLOAT_64 = "float64"
    UINT_32 = "uint32"


class DatasetTable(TypedDict):
    name: npt.NDArray[np.str_]
    hdf5_type: Sequence[PandaHdf5DatasetType]


class SeqTrigger(str, Enum):
    IMMEDIATE = "Immediate"
    BITA_0 = "BITA=0"
    BITA_1 = "BITA=1"
    BITB_0 = "BITB=0"
    BITB_1 = "BITB=1"
    BITC_0 = "BITC=0"
    BITC_1 = "BITC=1"
    POSA_GT = "POSA>=POSITION"
    POSA_LT = "POSA<=POSITION"
    POSB_GT = "POSB>=POSITION"
    POSB_LT = "POSB<=POSITION"
    POSC_GT = "POSC>=POSITION"
    POSC_LT = "POSC<=POSITION"


PydanticNp1DArrayInt32 = Annotated[
    np.ndarray[tuple[int], np.dtype[np.int32]],
    NpArrayPydanticAnnotation.factory(
        data_type=np.int32, dimensions=1, strict_data_typing=False
    ),
    Field(default_factory=lambda: np.array([], np.int32)),
]
PydanticNp1DArrayBool = Annotated[
    np.ndarray[tuple[int], np.dtype[np.bool_]],
    NpArrayPydanticAnnotation.factory(
        data_type=np.bool_, dimensions=1, strict_data_typing=False
    ),
    Field(default_factory=lambda: np.array([], dtype=np.bool_)),
]
TriggerStr = Annotated[Sequence[SeqTrigger], Field(default_factory=list)]


class SeqTable(Table):
    repeats: PydanticNp1DArrayInt32
    trigger: TriggerStr
    position: PydanticNp1DArrayInt32
    time1: PydanticNp1DArrayInt32
    outa1: PydanticNp1DArrayBool
    outb1: PydanticNp1DArrayBool
    outc1: PydanticNp1DArrayBool
    outd1: PydanticNp1DArrayBool
    oute1: PydanticNp1DArrayBool
    outf1: PydanticNp1DArrayBool
    time2: PydanticNp1DArrayInt32
    outa2: PydanticNp1DArrayBool
    outb2: PydanticNp1DArrayBool
    outc2: PydanticNp1DArrayBool
    outd2: PydanticNp1DArrayBool
    oute2: PydanticNp1DArrayBool
    outf2: PydanticNp1DArrayBool

    @classmethod
    def row(  # type: ignore
        cls,
        *,
        repeats: int = 1,
        trigger: str = SeqTrigger.IMMEDIATE,
        position: int = 0,
        time1: int = 0,
        outa1: bool = False,
        outb1: bool = False,
        outc1: bool = False,
        outd1: bool = False,
        oute1: bool = False,
        outf1: bool = False,
        time2: int = 0,
        outa2: bool = False,
        outb2: bool = False,
        outc2: bool = False,
        outd2: bool = False,
        oute2: bool = False,
        outf2: bool = False,
    ) -> "SeqTable":
        return Table.row(**locals())

    @model_validator(mode="after")
    def validate_max_length(self) -> "SeqTable":
        """
        Used to check max_length. Unfortunately trying the `max_length` arg in
        the pydantic field doesn't work
        """

        first_length = len(next(iter(self))[1])
        assert 0 <= first_length < 4096, f"Length {first_length} not in range."
        return self
